package cologne.eck.tools;

import java.awt.Component;
import java.awt.Window;
import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.ResourceBundle;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;
import java.util.zip.ZipInputStream;
import java.util.zip.ZipOutputStream;

import cologne.eck.all_peas.control.PeaControl;
import cologne.eck.all_peas.data.PeaProperties;
import cologne.eck.all_peas.gui.CursorManager;
import cologne.eck.all_peas.gui.PeaDialog;
import cologne.eck.all_peas.files.ExecutionTimeObserver;
import cologne.eck.all_peas.files.FileTypePanel;



public class ZipStuff {

	
	private static final int BUFFER_SIZE = 1024 * 1024;//8192;
	
	/**
	 * Unzip a zip archive and store included files in a selected folder
	 * 
	 * @param zipFileName		the archive to unzip
	 * @param folderNameToStore	the name of the folder to store the zip content
	 * 							this folder must exists
	 * @param window			the parent window
	 * @param bundle			the language bundle
	 * 
	 * @return	an array of created files from the zip archive
	 */
	public static File[] unzip(String zipFileName, String folderNameToStore, Window window, ResourceBundle bundle) {

		//System.out.println("zipFileName: " + zipFileName + ", folder: " + folderNameToStore);
		File zipFile = new File(zipFileName);
		if (folderNameToStore == "") {
			// files with absolute paths...        	
		} else { // check directory
			File folderToStore = new File(folderNameToStore);
			if ((! folderToStore.exists() )|| (! folderToStore.isDirectory())) {
				PeaDialog.showMessage(window, bundle.getString("select_folder_for_unzipped_files"), "Zip", 0);
				return null;
			}
			if (! zipFile.exists() || ! zipFile.isFile() || ! zipFile.canRead() || ! zipFile.canWrite() ) {
				PeaDialog.showMessage(window, bundle.getString("no_access") + "\n" + zipFile.getAbsolutePath(), "Zip", 0);
				return null;
			}
			folderNameToStore = folderNameToStore + File.separator;
		}
		// the list of files from zip
		ArrayList<File> list = new ArrayList<File>();
		FileInputStream fis = null;
		//buffer for read and write data to file
		byte[] buffer = new byte[1024];

		try {
			fis = new FileInputStream(zipFile);
		} catch (FileNotFoundException e) {
			PeaDialog.showMessage(window, 
					bundle.getString("unzip_failed") + 
					":\n"+ bundle.getString("file_not_found") + 
					":\n" + zipFileName, "Zip", 0);
			return null;
		}

		try {
			// set the wait cursor: 
			window.setCursor(CursorManager.getWaitCursor());

			ZipInputStream zis = new ZipInputStream(fis);
			// if first entry fails, break unzip operation with error message: 
			ZipEntry entry = null;
			try {

				entry = zis.getNextEntry();
			} catch (IOException e) {
				PeaDialog.showMessage(window, 
						bundle.getString("unzip_failed") + 
						":\n"+ bundle.getString("read_write_failed") + 
						":\n" + zipFileName, "Zip", 0);
				try {
					fis.close();
				} catch (IOException e1) {
					System.err.println(e1.toString() + ", " + e1.getLocalizedMessage());
				}
				return null;
			} 
			FileTypePanel ftp = (FileTypePanel)PeaProperties.getTypePanel();
			while(entry != null){
				String fileName = entry.getName();
				// ask to overwrite existing files
				if (PeaProperties.isOverwriteExistingFiles() == false) {
					if ( ftp.askToOverwriteFile(new File(folderNameToStore + fileName),true) == false){
						try {
							entry = zis.getNextEntry();
						} catch (IOException e) { // should never happen
							PeaDialog.showMessage(window, 
									bundle.getString("unzip_failed") + 
									":\n"+ bundle.getString("read_write_failed") + 
									":\n" + zipFileName, "Zip", 0);
							System.err.println(e.toString() + ", " + e.getLocalizedMessage());
							try {
								zis.close();
								fis.close();
							} catch (IOException e1) {
								System.err.println(e1.toString() + ", " + e1.getLocalizedMessage());
							}
							return null;							
						} 
						continue;
					}
				}
				File newFile = new File(folderNameToStore + fileName);
				/*		if (ftp.getFileComposer().getFileModel().containsFile(newFile)) {
				System.out.println("File already listed: " + newFile.getAbsolutePath());
			} else {*/
				list.add(newFile);
				//		}
				//create directories for sub directories in zip
				new File(newFile.getParent()).mkdirs();
				FileOutputStream fos;
				// if one file fails, continue with next file and show error message
				// do not break unzip operation
				try {
					fos = new FileOutputStream(newFile);
					int len;
					while ((len = zis.read(buffer)) > 0) {
						fos.write(buffer, 0, len);
					}
					fos.close();
					zis.closeEntry();

				} catch (FileNotFoundException e) {
					PeaDialog.showMessage(window, 
							bundle.getString("unzip_failed") + 
							":\n"+ bundle.getString("file_not_found") + 
							":\n" + folderNameToStore + fileName, "Zip", 0);

				} catch (IOException e) {
					PeaDialog.showMessage(window, 
							bundle.getString("unzip_failed") + 
							":\n"+ bundle.getString("read_write_failed") + 
							":\n" + folderNameToStore + fileName, "Zip", 0);
				} finally {
					// try next entry
					try {
						entry = zis.getNextEntry();
						continue;
					} catch (IOException e1) {
						System.err.println(e1.toString() + ", " + e1.getLocalizedMessage());
					}
				}
			}
			//close last ZipEntry
			try {
				zis.closeEntry();
				zis.close();
				fis.close();
			} catch (IOException e) {
				System.err.println(e.toString() + ", " + e.getLocalizedMessage());
			}

			File[] fileArray = new File[list.size()];
			return list.toArray(fileArray);
		} finally {
			window.setCursor(CursorManager.getDefaultCursor());
		}
	}


	

	/**
	 * Creates a zip file containing all chosen files
	 * 
	 * @param fileNames			file names to zip
	 * @param includeFDolders	true: include folders, false: ignore folders
	 * @param includeChildren	include children of folders - Note: for false, the structure remains,
	 * 							for true: all children will be included,even if not listed yet
	 * @param zipFileName		the name of the zip file to create
	 * @param zipFolder			the folder to store the zip file to create
	 * @param component			parent component for messages
	 * @param bundle			ResourceBundle to us
	 * 
	 * @return		null if the function performs successfully, 
	 * 				an error message if the function fails
	 */
	public static String zipFiles(String[] fileNames, boolean includeFolders, boolean includeChildren, 
				String zipFileName,  File zipFolder, Component component, ResourceBundle bundle) {

		if (PeaProperties.isOverwriteExistingFiles() == false) {
			if ( ( (FileTypePanel) ( PeaProperties.getTypePanel() ) ).askToOverwriteFile(new File( zipFolder + File.separator + zipFileName), true) == false){
				System.out.println("Creating zip file canceled");
				return bundle.getString("cancel") + ":\n " +zipFileName;
			}
		}
	    try {
	    	// set the wait cursor: 
			component.setCursor(CursorManager.getWaitCursor());
			
	    	//BufferedInputStream origin = null;
	        FileOutputStream dest = new FileOutputStream(zipFolder + File.separator + zipFileName);
	        // target zip output stream
	        ZipOutputStream out = new ZipOutputStream(new BufferedOutputStream(dest));
	        // compress:
	        out.setMethod(ZipOutputStream.DEFLATED);
	        for (int i=0; i<fileNames.length; i++) {
	        	addDirOrFileToZip(out, new File (fileNames[i]), includeFolders, includeChildren,
	        			fileNames[i].substring(0, fileNames[i].lastIndexOf(File.separator)), 
	        			0);
	        }
	        out.close();
	    } catch (FileNotFoundException e) {
	    	//e.printStackTrace();
			System.err.println("ZipStuff (zipFolder): " + e.toString() + ", " + e.getMessage());			
			return e.toString() + ", " + e.getMessage();
		} catch (IOException e) {
	    	//e.printStackTrace();
			System.err.println("ZipStuff (zipFolder): " + e.toString() + ", " + e.getMessage());	
			return e.toString() + ", " + e.getMessage();
		} catch (Exception e) {
	    	//e.printStackTrace();
			System.err.println("ZipStuff (zipFolder): " + e.toString() + ", " + e.getMessage());			
			return e.toString() + ", " + e.getMessage();		
		} finally {
			component.setCursor(CursorManager.getDefaultCursor());
		}
		return null;
	}
	

	/**
	 * Creates a zip file containing all files of the given folder, but not
	 * sub-directories and files inside of sub-directories. 
	 * The zip file is stored beside the folder. 
	 * 
	 * @param folderName	the directory name which contains the files to archive
	 * @param zipName		name of the zip file without extension ".zip"
	 * 
	 * @return			null if the function performs successfully, 
	 * 					an error message if the function fails
	 */
	public static String zipFolder(String folderName, String zipName, Window window, ResourceBundle bundle) {
		
		// partial error, but not complete failure of zip creation
		String partialError = "";
		File folder = new File(folderName);
		if (! folder.exists()) {
			System.err.println("ZipStuff (zipFolder): directory to zip does not exist");
			return "ZipStuff (zipFolder): directory to zip does not exist";
		}
		if (! folder.canRead()){
			System.err.println("ZipStuff (zipFolder): can not read directory to zip");
			return "ZipStuff (zipFolder): can not read directory to zip";		
		}
		if (! folder.isDirectory()){
			System.err.println("ZipStuff (zipFolder): file to zip is not directory");
			return "ZipStuff (zipFolder): file to zip is not directory";
		}
		if (FileTools.getNumberAndSize(folderName)[1] > 1024*1024*128) { // > 128MB
			if (ExecutionTimeObserver.showWarningAndBreakOption(
		    		PeaControl.getDialog().getLockFrame(), PeaProperties.getBundle().getString("size_warning"), false) 
					== false) {
				return "Folder content is too large";
			}
		}
		String path = folderName.substring(0, folderName.lastIndexOf(File.separator));
		String zipFileName = path + File.separator + zipName + ".zip";
		
		if (PeaProperties.isOverwriteExistingFiles() == false) {
			if ( ( (FileTypePanel) ( PeaProperties.getTypePanel() ) ).askToOverwriteFile(new File( zipFileName), true) == false){
				System.out.println("Creating zip file canceled");
				return bundle.getString("cancel") + ":\n " +zipFileName;
			}
		}

	    try {
	    	// set the wait cursor: 
			window.setCursor(CursorManager.getWaitCursor());

	        FileOutputStream dest = new FileOutputStream(zipFileName);
	        // target zip output stream
	        ZipOutputStream out = new ZipOutputStream(new BufferedOutputStream(dest));
	        // compress:
	        out.setMethod(ZipOutputStream.DEFLATED);
	        // get files of folder:
	        File[] filesOfFolder = folder.listFiles();
	        // index to remove absolute part of path
            int substringIndex = folder.getAbsolutePath().length();
	        int len = filesOfFolder.length;
	        for (int i=0; i<len; i++) {
		        String fileName = filesOfFolder[i].getAbsolutePath();
	            String error = addDirOrFileToZip(out, filesOfFolder[i], true, true,
	            	fileName.substring(0, fileName.lastIndexOf(File.separator)), 
	            	substringIndex);
	            if (error != null) {
	            	System.err.println("Adding file to zip failed: " + fileName + "\n" + error);
	            	partialError += "\n" + error + ": " + fileName;
	            }
	        }

	        out.flush();
	        out.close();
	        dest.flush();
	        dest.close();
	    } catch (FileNotFoundException e) {
	    	//e.printStackTrace();
			System.err.println("ZipStuff (zipFolder): " + e.toString() + ", " + e.getMessage());
			e.printStackTrace();
			return e.toString() + ", " + e.getMessage();
		} catch (IOException e) {
	    	//e.printStackTrace();
			System.err.println("ZipStuff (zipFolder): " + e.toString() + ", " + e.getMessage());	
			return e.toString() + ", " + e.getMessage();
		} catch (Exception e) {
	    	//e.printStackTrace();
			System.err.println("ZipStuff (zipFolder): " + e.toString() + ", " + e.getMessage());
			e.printStackTrace();
			return e.toString() + ", " + e.getMessage();		
		} finally {
			window.setCursor(CursorManager.getDefaultCursor());
		}
	    // Show if only adding some files failed:
	    if (! partialError.equals("")) {
	    	PeaDialog.showMessage(window, bundle.getString("partial_zip_error") +":\n" + partialError, "Zip", 0);
	    }
	    return null;
	}
	/**
	 * Add a file or directory to a zip file. The structure of the
	 * directory remains. 
	 * 
	 * @param zos					ZipOutputStream
	 * @param fileToZip				file to add to the zip archive
	 * @param includeFolders		true: include folders,false: ignore folders
	 * @param includeChildren		true: include children of folders, even if files are not listed
	 * 								false: include folder only, but no children
	 * @param parentDirectoryName	the parent directory of the file to add
	 * @param substringIndex		index to substring to transform
	 * 								an absolute to a relative path
	 * @return	errorMessage or null
	 */
	private static String addDirOrFileToZip(ZipOutputStream zos, File fileToZip, 
			boolean includeFolders,
			boolean includeChildren,
			String parentDirectoryName, int substringIndex)  {
	    if (fileToZip == null || !fileToZip.exists() || !fileToZip.canRead()) {
	    	System.err.println("Not a valid file to zip: " + fileToZip);
	        return ("Not a valid file to zip: " + fileToZip);
	    }
	    String zipEntryName = fileToZip.getName();
	    if (parentDirectoryName!=null && !parentDirectoryName.isEmpty()) {
	    	// Add a slash to mark as directory
	    	// remove absolute part of path
	        zipEntryName = parentDirectoryName.substring(
	        		substringIndex, parentDirectoryName.length()) + "/" + fileToZip.getName();
	    }
	    if (fileToZip.isDirectory()) {
	    	if (includeFolders == true) {
	    		if (includeChildren == true) {
	    			// execution time warning for size
	    			if (FileTools.getNumberAndSize(fileToZip.getAbsolutePath())[1] > 1024*1024*128) { // > 256MB
	    				if (ExecutionTimeObserver.showWarningAndBreakOption(
	    						null, PeaProperties.getBundle().getString("size_warning"), false) 
	    						== true) {
	    					for (File file : fileToZip.listFiles()) {
	    						addDirOrFileToZip(zos, file, includeFolders, includeChildren, zipEntryName, 0);
	    					}
	    				}
	    			} else {
	    				for (File file : fileToZip.listFiles()) {
	    					addDirOrFileToZip(zos, file, includeFolders, includeChildren, zipEntryName, 0);
	    				}
	    			}
	    		}
	    	} else {
	    		return null;
	    	}

	    } else {
	        byte[] buffer = new byte[BUFFER_SIZE];
	        FileInputStream fis;
			try {
				fis = new FileInputStream(fileToZip);
		        zos.putNextEntry(new ZipEntry(zipEntryName));
		        int length;
		        while ((length = fis.read(buffer)) > 0) {
		            zos.write(buffer, 0, length);
		        }
		        zos.closeEntry();
		        fis.close();
			} catch (FileNotFoundException e) {
				System.err.println("ZipStuff (adding): " + e.toString() + ", " + e.getMessage());			
				return e.toString() + ", " + e.getMessage();	
			} catch (IOException e) {
				System.err.println("ZipStuff (adding): " + e.toString() + ", " + e.getMessage());			
				return e.toString() + ", " + e.getMessage();	
			} catch (Exception e) {
				System.err.println("ZipStuff (adding): " + e.toString() + ", " + e.getMessage());			
				return e.toString() + ", " + e.getMessage();	
			}
	    }
	    return null;
	}
	/**
	 * If addFilesToZip failed, error message contains the
	 * description of the error
	 */
	private static String errorMessage;
	/**
	 * Extract one file from the given zip file
	 * 
	 * @param fileName		name of the file to extract. 
	 * @param zipName		name of the zip file which contains the 
	 * 						file to extract
	 * 
	 * @return			the content of the extracted file as an 
	 * 					array of bytes
	 */
	public static byte[] getFileFromZipFile(String fileName, String zipName) {
		
		errorMessage = null;
		
		File zipFile = new File(zipName);
		if (! zipFile.exists()) {
			System.err.println("ZipStuff (getFileFromZip): zip file does not exist");
			errorMessage = "ZipStuff (getFileFromZip): zip file does not exist";
			return null;
		}
		if (! zipFile.canRead()){
			System.err.println("ZipStuff (getFileFromZip): can not read zip file");
			errorMessage = "ZipStuff (getFileFromZip): can not read zip file";
			return null;
		}
		if (zipFile.isDirectory()){
			System.err.println("ZipStuff (getFileFromZip): zip file is directory");
			errorMessage = "ZipStuff (getFileFromZip): zip file is directory";
			return null;
		}

        FileInputStream fis = null;
		try {
			fis = new FileInputStream(zipFile);
	         ZipInputStream zis = new ZipInputStream(new BufferedInputStream(fis));
	         ZipEntry entry;
	         while((entry = zis.getNextEntry()) != null) {
	            //System.out.println("Extracting: " +entry);
	        	 
	        	 if (entry.getName().endsWith(fileName)){
	 	            //FileOutputStream fos = new FileOutputStream(entry.getName());
	        		 ByteArrayOutputStream buffOut = new ByteArrayOutputStream(BUFFER_SIZE);
	        		 byte[] buffer = new byte[BUFFER_SIZE];
	                 int bytesRead = 0;
	                 while ((bytesRead = zis.read(buffer)) != -1) {
	                     buffOut.write(buffer, 0, bytesRead);
	                 }		        		        		 
	        		 return (buffOut == null) ? null : buffOut.toByteArray();
	        	 }
	         }
	         fis.close();
		} catch (FileNotFoundException e) {
			System.err.println("ZipStuff (getFileFromZip): " + e.toString() + ", " + e.getMessage());
			errorMessage =  e.toString() + ", " + e.getMessage();
			//e.printStackTrace();
		} catch (IOException e) {
			System.err.println("ZipStuff (getFileFromZip): " + e.toString() + ", " + e.getMessage());
			errorMessage = e.toString() + ", " + e.getMessage();
			//e.printStackTrace();
		} catch (Exception e) {
			System.err.println("ZipStuff (getFileFromZip): " + e.toString() + ", " + e.getMessage());
			errorMessage = e.toString() + ", " + e.getMessage();
			//e.printStackTrace();
		}
		return null;
	}
	
	/**
	 * Add one file to a given zip file. 
	 * 1. The original zip file is unzipped, 
	 * 2. the entries are added to a temporary zip file called tmp.zip, 
	 * 3. the file is appended to the temporary zip file, 
	 * 4. the given original zip file is deleted 
	 * 5. and the temporary zip file is renamed to the original zip file. 
	 * 
	 * 
	 * @param fileName
	 * @param oldZipName
	 * 
	 * @return an error message if the operation failed, null otherwise
	 */
	public static String addFileToZip(String fileName, String oldZipName) {	
		
		String errorMessage = null;
			
	       // read war.zip and write to append.zip
	     ZipFile zipFile;
			try {
				zipFile = new ZipFile(oldZipName);
		        ZipOutputStream zipStream = new ZipOutputStream(new FileOutputStream("test/tmp.zip"));

		        // copy files from existing zip
		        Enumeration<? extends ZipEntry> entries = zipFile.entries();
		        while (entries.hasMoreElements()) {
		            ZipEntry entry = entries.nextElement();
		            byte[] buffer = new byte[BUFFER_SIZE];
		           // System.out.println( entry.getName());
		            zipStream.putNextEntry(entry);
		            if (!entry.isDirectory()) {
		            	InputStream input = zipFile.getInputStream(entry);
		                int bytesRead;
		                while ((bytesRead = input.read(buffer))!= -1) {
		                    zipStream.write(buffer, 0, bytesRead);
		                }
		            }
		        }
		        // append file
		        ZipEntry entry = new ZipEntry(fileName);
		        zipStream.putNextEntry(entry);
		        zipStream.write(ReadResources.readExternFile("test" + File.separator +  entry.getName()));
		        zipStream.closeEntry();

		        // close
		        zipFile.close();
		        zipStream.close();
			} catch (FileNotFoundException e) {
				System.err.println("ZipStuff (addFileToZip): " + e.toString() + ", " + e.getMessage());
				errorMessage =  e.toString() + ", " + e.getMessage();
				//e.printStackTrace();
			} catch (IOException e) {
				System.err.println("ZipStuff (addFileToZip): " + e.toString() + ", " + e.getMessage());	
				errorMessage =  e.toString() + ", " + e.getMessage();
				//e.printStackTrace();
			} catch (Exception e) {
				System.err.println("ZipStuff (addFileToZip): " + e.toString() + ", " + e.getMessage());	
				errorMessage =  e.toString() + ", " + e.getMessage();
				//e.printStackTrace();
			}
			
			try {
			    File oldFile = new File(oldZipName);	    
			    oldFile.delete();		    
			    File tmpFile = new File("test/tmp.zip");
			    tmpFile.renameTo(oldFile);
			} catch (Exception e) {
				errorMessage =  "Rename file failed: " + e.toString() + ", " + e.getMessage();
			}		    		    
		    return errorMessage;
		}

		/**
		 * If the function addFileToZip() failed,
		 * the error message contains a description
		 * of the error
		 * 
		 * @return the errorMessage
		 */
		public static String getErrorMessage() {
			return errorMessage;
		}
}
